package at.ac.univie.cs.swa.component.architectureabstraction.utils

import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.AggregationCascadeAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ArchitectureAbstractionDSLFactory
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.CallbackAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentConnector
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentDef
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.CompositeCascadeAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ConnectorAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.Expression
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.GroupingAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.IndirectionAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.LayeringAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.PatternInstancePrimitiveTarget
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.PrimitiveAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ShieldAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.SimpleConnectorAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.Transformation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.TypingAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.VirtualConnectorAnnotation
import at.ac.univie.cs.swa.pattern.catalog.PatternPrimitiveUtils
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.AggregationCascade
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Callback
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Component
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.CompositeCascade
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Grouping
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Indirection
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Layering
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PRIMITIVE_TYPE
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Primitive
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Shield
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.SimpleRelation
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Typing
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.VirtualConnector
import java.util.Collections
import java.util.HashSet
import java.util.List
import java.util.Set
import org.eclipse.core.resources.ResourcesPlugin
import org.eclipse.emf.common.util.URI
import org.eclipse.emf.ecore.EObject
import org.eclipse.emf.ecore.resource.Resource
import org.eclipse.emf.ecore.util.EcoreUtil$EqualityHelper
import org.eclipse.uml2.uml.Class
import org.eclipse.uml2.uml.Classifier
import org.eclipse.uml2.uml.Interface
import org.eclipse.uml2.uml.Package
import org.eclipse.uml2.uml.Relationship
import org.eclipse.uml2.uml.Usage
import java.util.LinkedList

class ArchitectureAbstractionUtils extends PatternPrimitiveUtils {

	protected static val EqualityHelper equalityHelper = new EqualityHelper();

	private static var ArchitectureAbstractionUtils = new ArchitectureAbstractionUtils

	def getGeneratorResourceURI(Resource r) {
		var uri = URI::createURI(r.URI.toPlatformString(false) + "appended");
		return uri
	}

	private Set<Relationship> relationshipCache = new HashSet<Relationship>

	def static Set<Relationship> myrelationships(Package p) {
		val relationShips = p.ownedElements.filter(typeof(Relationship)).toSet
		val childRelationShips = p.ownedElements.filter(typeof(Package)).map(e|myrelationships(e)).flatten
		relationShips.addAll(
			childRelationShips
		)

		return relationShips.toSet
	}

	def Set<ComponentDef> getLayerComponents(List<EObject> list, String layerName) {
		for (EObject eObj : list) {
			return handleTransformation(eObj, layerName)
		}
		return emptySet;
	}

	def private dispatch Set<ComponentDef> handleTransformation(EObject object, String layerName) {
		return emptySet
	}

	def private dispatch Set<ComponentDef> handleTransformation(Transformation transformation, String layerName) {
		val result = transformation.components.filter(e|e.annotations.exists(f|isInLayer(f, layerName)))
		return result.toSet
	}

	def relatedTargetClassifiers(Classifier c) {
		return relatedTargetClassifiers(c, false)
	}

	def relatedTargetClassifiers(Classifier c, boolean print) {
		val propertyTypes = new LinkedList<Classifier>
		if (c instanceof Class) {
			val claz = c as Class
			val props = claz.ownedAttributes

			for (p : props) {
				val type = p.type
				if(type instanceof Classifier) {
					val classifier = type as Classifier
					propertyTypes.add(classifier)
				}
				//propertyTypes.add(p.type.)
			}

		}
		val allRelationShips = c.nearestPackage.allRelationships
		val usageRelationShips = allRelationShips.filter(typeof(Usage))
		val relationships = usageRelationShips.filter(e|e.clients.map(f|f.name).toSet.contains(c.name));
		val relatedClassifiers = relationships.map(e|e.suppliers).flatten.filter(typeof(Classifier)).toSet
		relatedClassifiers.addAll(propertyTypes)
		return relatedClassifiers
	}

	def getIFile(EObject eObject) {
		val Resource eResource = eObject.eResource();
		val URI eUri = eResource.getURI();
		if (eUri.isPlatformResource()) {
			val String platformString = eUri.toPlatformString(true);
			return ResourcesPlugin::workspace.root.findMember(platformString)
		}
		return null
	}

	/*
	 * this cache does not check if the resources are changed underneath ....
	 */
	def Set<Relationship> allRelationships(Package p) {
		if (!relationshipCache.empty) {
			return relationshipCache
		}
		if (p.owner != null && p.owner instanceof Package) {
			val owner = p.owner as Package
			return allRelationships(owner)
		}
		val result = myrelationships(p)
		relationshipCache = result
		return result
	}

	//
	//	def Set<ComponentDef> targets(ConnectorAnnotation in) {
	//		var Set<ComponentDef> set = new HashSet<ComponentDef>
	//		set.add(in.target)
	//		return set
	//	}
	def Set<ComponentDef> calculatedDirectRelatedComponents(ComponentDef component, AbstractEvaluator evaluator) {
		val Set<Classifier> realizingClassifiers = evaluator.evaluateComponent(component).toSet
		val Set<Classifier> relatedClassifiers = realizingClassifiers.map(c|c.relatedTargetClassifiers).flatten.toSet
		val Set<ComponentDef> relatedComponents = relatedClassifiers.map(c|evaluator.getComponentForClassifier(c)).
			filterNull.toSet
		return relatedComponents;
	}
	
	def Set<ComponentDef> componentLevelDefinedDirectRelatedComponents(ComponentDef component) {
		return component.connectors.map[e|e.targetComponents].flatten.toSet
	}

	def Set<Classifier> relationClassifiers(ComponentDef c1, ComponentDef c2, AbstractEvaluator evaluator) {
		val Set<Classifier> realizingClassifiers = evaluator.evaluateComponent(c2).toSet
		val Set<Classifier> relatedClassifiers = realizingClassifiers.map(c|c.relatedTargetClassifiers).flatten.
			filterNull.toSet
		return relatedClassifiers.filter(c|evaluator.getComponentForClassifier(c)?.equals(c1)).filterNull.toSet
	}

	def Expression createConnectorExpressionFor(Set<Classifier> classifiers) {

		if (classifiers.size() > 1) {
			val hashedClassifiers = new HashSet(classifiers)
			return createExpressionForClassifiers(hashedClassifiers)
		} else {
			return classifiers.head.createExpressionFor
		}
	}

	def Expression createExpressionForClassifiers(Set<Classifier> classifiers) {
		if (classifiers.size() > 1) {
			val expr = ArchitectureAbstractionDSLFactory::eINSTANCE.createOrComposition;
			expr.setLeft(createExpressionFor(classifiers.head))
			expr.setRight(createExpressionForClassifiers(classifiers.tail.toSet))
			return expr;
		}
		return createExpressionFor(classifiers.head)
	}

	def dispatch Expression createExpressionFor(Classifier c) {
		throw new RuntimeException("This code should never be executed!")
	}

	def dispatch Expression createExpressionFor(Class c) {
		val expr = ArchitectureAbstractionDSLFactory::eINSTANCE.createIsClass;
		expr.setRelatedTo(c);
		return expr;
	}

	def dispatch Expression createExpressionFor(Interface c) {
		val expr = ArchitectureAbstractionDSLFactory::eINSTANCE.createSpecificInterface;
		expr.setRelatedTo(c);
		return expr;
	}

	def static Set<ComponentDef> targetComps(ComponentConnector c) {
		return ArchitectureAbstractionUtils.targetComponents(c)
	}

	def Set<ComponentDef> targetComponents(ComponentConnector c) {
		val allTargets = c.targets
		return allTargets.map[e|e.resolve].flatten.toSet
	}

	def dispatch Set<ComponentDef> resolve(PatternInstancePrimitiveTarget target) {
		throw new RuntimeException("Missing resolve method for type:" + target.eClass.name)
	}

	def dispatch Set<ComponentDef> resolve(ComponentDef target) {
		val set = new HashSet
		set.add(target)
		return set
	}

	def dispatch Set<ComponentDef> resolve(GroupingAnnotation targetGroup) {
		val set = new HashSet<ComponentDef>
		set.addAll(targetGroup.componentsInGroup)
		return set
	}

	def Transformation containingTransformation(EObject e) {
		if (e.eContainer != null) {
			if (e.eContainer instanceof Transformation) {
				return e.eContainer as Transformation
			} else {
				return e.eContainer.containingTransformation
			}
		} else {
			return null
		}
	}

	def Set<? extends PatternInstancePrimitiveTarget> targets(ComponentConnector c) {
		return c.annotation.annotationTargets
	}

	def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(ConnectorAnnotation ca) {
		throw new RuntimeException("Missing annotationTargets method for type:" + ca.eClass.name)

		//return Collections::emptySet
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(IndirectionAnnotation ca) {
			return ca.targets.toSet
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(CallbackAnnotation ca) {
			val set = new HashSet
			set.add(ca.target)
			return set
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(ShieldAnnotation ca) {
			return ca.targets.toSet
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(SimpleConnectorAnnotation ca) {
			return ca.targets.toSet
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(TypingAnnotation ca) {
			return ca.targets.toSet
		}

		def dispatch Set<? extends PatternInstancePrimitiveTarget> annotationTargets(VirtualConnectorAnnotation ca) {
			return ca.targets.toSet
		}

		def static dispatch boolean isInLayer(PrimitiveAnnotation annotation, String layerName) {
			return false
		}

		def static dispatch boolean isInLayer(LayeringAnnotation annotation, String layerName) {
			return isInGroup(annotation, layerName)
		}

		def static dispatch boolean isInGroup(PrimitiveAnnotation annotation, String groupId) {
			return false
		}

		def static dispatch boolean isInGroup(GroupingAnnotation annotation, String groupId) {
			return annotation.groupId.equals(groupId)
		}

		def Set<ComponentDef> componentsInGroup(List<ComponentDef> components, String groupId) {
			return components.filter(
				c|
					c.annotations.exists(
						e|e instanceof GroupingAnnotation && (e as GroupingAnnotation).groupId.equals(groupId))).toSet
		}

		def Set<ComponentDef> componentsInGroup(GroupingAnnotation annotation) {
			val Transformation t = annotation.containingTransformation
			if (t == null) {
				return Collections::emptySet
			} else {
				return componentsInGroup(t.components, annotation.groupId)
			}
		}

		def dispatch PRIMITIVE_TYPE getType(PrimitiveAnnotation pa) {
			throw new RuntimeException("Missing implementation of getType for " + pa.eClass.name)
		}

		def dispatch PRIMITIVE_TYPE getType(SimpleConnectorAnnotation pa) {
			PRIMITIVE_TYPE::SIMPLE_RELATION_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(SimpleRelation pa) {
			PRIMITIVE_TYPE::SIMPLE_RELATION_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(AggregationCascadeAnnotation pa) {
			PRIMITIVE_TYPE::AGGREGATION_CASCADE_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(AggregationCascade pa) {
			PRIMITIVE_TYPE::AGGREGATION_CASCADE_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(CompositeCascadeAnnotation pa) {
			PRIMITIVE_TYPE::COMPOSITE_CASCADE_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(CompositeCascade pa) {
			PRIMITIVE_TYPE::COMPOSITE_CASCADE_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(CallbackAnnotation pa) {
			PRIMITIVE_TYPE::CALLBACK_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Callback pa) {
			PRIMITIVE_TYPE::CALLBACK_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(IndirectionAnnotation pa) {
			PRIMITIVE_TYPE::INDIRECTION_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Indirection pa) {
			PRIMITIVE_TYPE::INDIRECTION_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(TypingAnnotation pa) {
			PRIMITIVE_TYPE::TYPING_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Typing pa) {
			PRIMITIVE_TYPE::TYPING_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(VirtualConnectorAnnotation pa) {
			PRIMITIVE_TYPE::VIRTUAL_CONNECTOR_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(VirtualConnector pa) {
			PRIMITIVE_TYPE::VIRTUAL_CONNECTOR_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Component pa) {
			PRIMITIVE_TYPE::COMPONENT_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(ShieldAnnotation pa) {
			PRIMITIVE_TYPE::SHIELD_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Shield pa) {
			PRIMITIVE_TYPE::SHIELD_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(GroupingAnnotation pa) {
			PRIMITIVE_TYPE::GROUPING_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Grouping pa) {
			PRIMITIVE_TYPE::GROUPING_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(LayeringAnnotation pa) {
			PRIMITIVE_TYPE::LAYERING_TYPE
		}

		def dispatch PRIMITIVE_TYPE getType(Layering pa) {
			PRIMITIVE_TYPE::LAYERING_TYPE
		}

		def boolean matches(PrimitiveAnnotation pa, Primitive p) {
			if (p instanceof Component) {
				return true
			}
			if (pa == null || pa.type == null || p.type == null) {
				throw new RuntimeException(
					'''could not determine primitive type for primitive "«p»" and primitive annotation «pa»''')
			}
			return pa.type.equals(p.type)
		}
		
		def dispatch targetName(PatternInstancePrimitiveTarget target) {
			throw new RuntimeException("method not implement for " + target.eClass)
		}
		
		def dispatch targetName(GroupingAnnotation target) {
			return target.groupId
		}
		
		def dispatch targetName(ComponentDef target) {
			return target.name
		}
		
	}
	