package at.ac.univie.cs.swa.component.architectureabstraction.utils

import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ArchitectureAbstractionDSLFactory
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentConnector
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentDef
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.Transformation
import java.util.Set
import org.eclipse.xtext.EcoreUtil2
import org.eclipse.xtext.scoping.IScopeProvider
import org.eclipse.uml2.uml.Classifier
import java.util.HashSet

class SimpleRelationGenerator {

	extension ArchitectureAbstractionUtils utils = new ArchitectureAbstractionUtils

	val IScopeProvider scopeProvider;

	static AssociationGeneratorEvaluator evaluator = new AssociationGeneratorEvaluator()

	new(IScopeProvider sp) {
		this.scopeProvider = sp;
		if (evaluator != null) {
			evaluator.setScopeProvider(scopeProvider)
		}
	}

	def public getCandidates(Transformation transformation) {

		// generate all instances at once and work from the cache from now on
		val candidates = new HashSet
		evaluator.transform(transformation, getGeneratorResourceURI(transformation.eResource))
		for (ComponentDef c : transformation.components) {
			val sourceRelatedComponents = calculatedDirectRelatedComponents(c, evaluator).filterNull
			val Set<? extends ComponentDef> describedRelatedComponents = c.connectors.map(e|e.targetComponents).flatten.
				toSet
			for (targetComponent : sourceRelatedComponents) {
				if (!describedRelatedComponents.contains(targetComponent) && !targetComponent.name.equals(c.name)) {
					val sourceClassifiers = evaluator.evaluateComponent(targetComponent).toSet
					sourceClassifiers.forEach[e|EcoreUtil2::resolveAll(e)]
					val classifiers = relationClassifiers(targetComponent, c, evaluator) // .filter[classifier|sourceClassifiers.contains(classifier)].toSet;
					if (!classifiers.isEmpty()) {

						//						createConnector(targetComponent, c, classifiers)
						candidates.add(createCandidate(targetComponent, c, classifiers))
					}
				}
			}
		}
		return candidates
	}

	def createCandidate(ComponentDef targetComponent, ComponentDef c, Set<Classifier> classifiers) {
		return new ConnectorCandidate(targetComponent, c, classifiers)
	}

	def createConnector(ConnectorCandidate cc) {
		createConnector(cc.target, cc.source, cc.classifiers, cc.generateImplementationExpression)
	}

	def private createConnector(ComponentDef targetComponent, ComponentDef c, Set<Classifier> classifiers,
		boolean generateImplementationExpression) {
		val ComponentConnector compConnector = ArchitectureAbstractionDSLFactory::eINSTANCE.createComponentConnector
		val annotation = ArchitectureAbstractionDSLFactory::eINSTANCE.createSimpleConnectorAnnotation

		//						annotation.setName("<Auto-generated>Please enter a name here!")
		annotation.targets.add(targetComponent)
		compConnector.annotation = annotation

		c.connectors.add(compConnector)
		if (generateImplementationExpression) {
			val expr = createConnectorExpressionFor(classifiers)
			compConnector.implementingExpression.add(expr)

		}
	}

}

class ConnectorCandidate {
	private var ComponentDef target;
	private var ComponentDef source;
	private var Set<Classifier> classifiers;
	private var boolean generateImplementationExpression;

	new(ComponentDef target, ComponentDef source, Set<Classifier> classifiers) {
		this.target = target;
		this.source = source;
		this.classifiers = classifiers;
	}

	def getTarget() {
		return target;
	}

	def getSource() {
		return source;
	}

	def getClassifiers() {
		return classifiers;
	}

	def isGenerateImplementationExpression() {
		return generateImplementationExpression
	}

	def void setGenerateImplementationExpression(boolean doGenerate) {
		generateImplementationExpression = doGenerate;
	}

	override toString() {
		return '''«source.name» --> «target.name»'''
	}
}
