package at.ac.univie.cs.swa.pattern.catalog.validation

import org.eclipse.xtext.validation.Check
import org.eclipse.xtext.validation.CheckType
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.CategoryOrderings
import org.eclipse.emf.ecore.EObject
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Catalog
import java.util.List
import java.util.LinkedList
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Category
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PatternCatalogDSLPackage
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Range
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PrimitiveCategory
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.ComponentPrimitive
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Shield
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Relation
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.RelationPrimitive
import at.ac.univie.cs.swa.pattern.catalog.PatternPrimitiveUtils

class PatternCatalogDSLValidator  extends
		AbstractPatternCatalogDSLJavaValidator {

	extension PatternPrimitiveUtils ppu = new PatternPrimitiveUtils

	@Check
	def public checkCategoryUsages(CategoryOrderings currentOrdering) {
		val EObject container = currentOrdering.eContainer();
		if (container instanceof Catalog) {
			val Catalog catalog = container as Catalog;
			val List<CategoryOrderings> orderings = catalog.getOrderings();
			for (CategoryOrderings ordering : orderings) {
				if (!ordering.equals(currentOrdering)) {
					val LinkedList<Category> doubleUsedCategories = new LinkedList<Category>(
							currentOrdering.getOrder());
					doubleUsedCategories.retainAll(ordering.getOrder());
					for (Category c : doubleUsedCategories) {
						error("Every Category can only be used in one Ordering: "
								+ c.getName(),
								currentOrdering,
								PatternCatalogDSLPackage::eINSTANCE.categoryOrderings_Order,
								currentOrdering.getOrder().indexOf(c));
					}
				}
			}
		}
	}

	@Check(CheckType::FAST)
	def public checkRange(Range r) {
		val String upperBoundString = r.getUpperBound();
		val String lowerBoundString = r.getLowerBound();
		if (lowerBoundString != null && upperBoundString != null
				&& !upperBoundString.equals("*")) {
			try {
				val int lowerBound = Integer::parseInt(lowerBoundString);
				val int upperBound = Integer::parseInt(upperBoundString);

				if (upperBound < lowerBound) {
					error("Upper bound cannot be lower than lower bound!",
							PatternCatalogDSLPackage::eINSTANCE.range_UpperBound);
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

	@Check(CheckType::FAST)
	def public checkShieldPrimitive(PrimitiveCategory pc) {
		val ComponentPrimitive primitive = pc.getPrimitive();
		if (primitive instanceof Shield) {
			var boolean found = searchRelations(pc);			
			if (!found) {
				error("Shield primitive requires a direct relation to at least 1 shielded object!",
						PatternCatalogDSLPackage::eINSTANCE.primitiveCategory_Relations);
			}

		}
	}
	
	def private searchRelations(PrimitiveCategory pc) {
			for (Relation r : pc.getRelations()) {
				val RelationPrimitive relationPrimitive = r.getPrimitive();
				if (relationPrimitive.isDirectRelation) {
					if (!pc.equals(r.getRelatedTarget())) {
						return true;
					} else {
						warning("Shield primitive cannot shield itself!",
								PatternCatalogDSLPackage::eINSTANCE.primitiveCategory_Relations);
					}
				}
			}
		return false;
	}
}