/*
 * generated by Xtext
 */
package at.ac.univie.cs.swa.pattern.scoping

import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentDef
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.GroupingAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.PrimitiveAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.utils.ArchitectureAbstractionUtils
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Component
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Grouping
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Primitive
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PrimitiveCategory
import at.ac.univie.cs.swa.pattern.instanceDSL.PatternInstance
import at.ac.univie.cs.swa.pattern.instanceDSL.PrimitiveToComponentMapping
import com.google.common.base.Predicate
import java.util.HashSet
import java.util.List
import java.util.Set
import org.eclipse.emf.ecore.EObject
import org.eclipse.emf.ecore.EReference
import org.eclipse.emf.ecore.resource.ResourceSet
import org.eclipse.xtext.EcoreUtil2
import org.eclipse.xtext.resource.IEObjectDescription
import org.eclipse.xtext.scoping.IScope
import org.eclipse.xtext.scoping.Scopes
import org.eclipse.xtext.scoping.impl.AbstractDeclarativeScopeProvider
import org.eclipse.xtext.scoping.impl.FilteringScope

/**
 * This class contains custom scoping description.
 * 
 * see : http://www.eclipse.org/Xtext/documentation.html#scoping
 * on how and when to use it 
 *
 */
class InstanceDSLScopeProvider extends AbstractDeclarativeScopeProvider {

	def public IScope scope_PrimitiveToComponentMapping_target(PrimitiveToComponentMapping mapping, EReference eRef) {
		debugPrint(eRef);
		val PrimitiveCategory primitivecategory = mapping.getPrimitivecategory();
		var Primitive primitive = primitivecategory.getPrimitive();
		primitive = ArchitectureAbstractionUtils::resolveProxy(primitive, primitivecategory);
		val IScope superscope = super.delegateGetScope(mapping, eRef);
//		val componentAnnotations = superscope.allElements.map[description|description.EObjectOrProxy].filter(typeof(ComponentDef)).map[c|c.annotations].flatten.toSet //.filter[typeof(ComponentDef)]
		var Predicate<IEObjectDescription> filter = null;
		if (primitive instanceof Grouping) {
			filter = new GroupingPredicate;
		} else if (primitive instanceof Component) {
			filter = new ComponentPredicate;
		} else {

			filter = new ComponentFilter(primitive, mapping.eResource().getResourceSet());
		}

		//		return superscope;
		return new FilteringScope(superscope, filter);
	}

	def public IScope scope_PrimitiveToComponentMapping_primitivecategory(PatternInstance patternInstance,
		EReference eRef) {
		debugPrint(eRef);
		val List<PrimitiveCategory> pcList = patternInstance.getPatternTemplate().getConsistsOf();
		val Set<PrimitiveCategory> primitives = new HashSet<PrimitiveCategory>();
		for (PrimitiveCategory pc : pcList) {
			primitives.add(pc);
		}

		//		final PrimitiveFilter filter = new PrimitiveFilter(primitives,
		//				patternInstance);
		//		return new FilteringScope(delegateGetScope(patternInstance, eRef),
		//				filter);
		return Scopes::scopeFor(primitives);

	}
	
	def private void debugPrint(EReference eRef) {
//		println(eRef);
	}

}

class ComponentFilter implements Predicate<IEObjectDescription> {
	private val Primitive primitive;
	private var ResourceSet resourceSet;
	private final ArchitectureAbstractionUtils utils = new ArchitectureAbstractionUtils();

	new(Primitive primitive, ResourceSet resourceSet) {
		this.primitive = primitive;
		this.resourceSet = resourceSet;

	}

	@Override
	override public boolean apply(IEObjectDescription input) {
		val EObject eObj = input.getEObjectOrProxy();
		if (input != null && eObj instanceof ComponentDef) {
			val ComponentDef compDef = EcoreUtil2::resolve(eObj, resourceSet) as ComponentDef;
			val List<PrimitiveAnnotation> annotations = compDef.getAnnotations();
			for (PrimitiveAnnotation annotation : annotations) {
				if (utils.matches(annotation, primitive))
					return true;

				//					if (helper.equals(primitive, primitive2)) {
				//						return true;
				//					}
				}
			}
			return false;
		}
	}

	class GroupingPredicate implements Predicate<IEObjectDescription> {
		override public boolean apply(IEObjectDescription input) {
			val eObjectOrProxy = input.getEObjectOrProxy()
//			println("xx:" + eObjectOrProxy.eClass.name)
			if (input != null && eObjectOrProxy instanceof GroupingAnnotation) {
				return true;
			}
			return false;
		}

	}

	class ComponentPredicate implements Predicate<IEObjectDescription> {
		override public boolean apply(IEObjectDescription input) {
			val eObjectOrProxy = input.getEObjectOrProxy()
			if (input != null && eObjectOrProxy instanceof ComponentDef) {
				return true;
			}
			return false;
		}
	}
	