/*
 * generated by Xtext
 */
package at.ac.univie.cs.swa.pattern.validation

//import org.eclipse.xtext.validation.Check

import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.ComponentDef
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.LayeringAnnotation
import at.ac.univie.cs.swa.component.architectureabstraction.architectureAbstractionDSL.PatternInstancePrimitiveTarget
import at.ac.univie.cs.swa.component.architectureabstraction.utils.ArchitectureAbstractionUtils
import at.ac.univie.cs.swa.component.architectureabstraction.utils.AssociationGeneratorEvaluator
import at.ac.univie.cs.swa.component.architectureabstraction.utils.RangeCheckUtilities
import at.ac.univie.cs.swa.component.architectureabstraction.utils.RangeCheckUtilitiesXtend
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Component
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.ComponentPrimitive
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Constraint
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Layering
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PatternTemplate
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.PrimitiveCategory
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Range
import at.ac.univie.cs.swa.pattern.instanceDSL.InstanceDSLPackage
import at.ac.univie.cs.swa.pattern.instanceDSL.Model
import at.ac.univie.cs.swa.pattern.instanceDSL.PatternInstance
import at.ac.univie.cs.swa.pattern.instanceDSL.PrimitiveToComponentMapping
import at.ac.univie.cs.swa.pattern.utils.XtendUtils
import com.google.inject.Inject
import java.util.Collections
import java.util.HashMap
import java.util.HashSet
import java.util.List
import java.util.Map
import java.util.Set
import org.eclipse.uml2.uml.Classifier
import org.eclipse.xtext.scoping.IScopeProvider
import org.eclipse.xtext.validation.Check
import at.ac.univie.cs.swa.pattern.catalog.patternCatalogDSL.Primitive
import org.eclipse.xtext.validation.CheckType

/**
 * Custom validation rules. 
 *
 * see http://www.eclipse.org/Xtext/documentation.html#validation
 */
class InstanceDSLValidator extends AbstractInstanceDSLValidator {

	extension RangeCheckUtilitiesXtend = new RangeCheckUtilitiesXtend

	extension ArchitectureAbstractionUtils = new ArchitectureAbstractionUtils

	private AssociationGeneratorEvaluator dslevaluator

	@Inject
	private IScopeProvider scopeProvider;

	def private AssociationGeneratorEvaluator getEvaluator() {
		if (dslevaluator == null) {
			dslevaluator = new AssociationGeneratorEvaluator();
			dslevaluator.setScopeProvider(scopeProvider);
		}

		// if(evaluator.getScopeProvider() == null) {
		// evaluator.setScopeProvider(scopeProvider);
		// }
		return dslevaluator;
	}

	new() {
		super();
	}
	
	
	@Check
	def public void checkAllAnnotatedPrimitivesExist(Model m) {
		val List<PatternInstance> patternInstances = m.getPatternInstances();
		for (PatternInstance patternInstance : patternInstances) {
			val PatternTemplate pattern = patternInstance.getPatternTemplate();
			val List<PrimitiveCategory> primitiveCategories = pattern
					.getConsistsOf();
			val Set<ComponentPrimitive> annotatedPrimitives = new HashSet<ComponentPrimitive>();
			for (PrimitiveCategory pc : primitiveCategories) {
				annotatedPrimitives.add(pc.getPrimitive());
			}
			val Set<Primitive> mapPrimitives = new HashSet<Primitive>();
			val List<PrimitiveToComponentMapping> mappings = patternInstance
					.getPrimitiveToComponentMappings();
			for (PrimitiveToComponentMapping mapping : mappings) {
				mapPrimitives.add(mapping.getPrimitivecategory().getPrimitive());
			}
			annotatedPrimitives.removeAll(mapPrimitives);
			// annotatedPrimitives.removeAll(mapPrimitives)
			if (annotatedPrimitives.size() > 0) {
				var String errorString = "The following primitives are definied in the pattern description but are not part of this pattern instance: ";
				var boolean firstRun = true;
				for (ComponentPrimitive p : annotatedPrimitives) {
					if (firstRun) {
						errorString = errorString + ", ";
						firstRun = false;
					}
					errorString = errorString + p.getClass().getSimpleName();
				}
				warning(errorString, patternInstance,
						XtendUtils::getPatternInstances_Template(), 0);
			}

		}
		
	}
	

	@Check
	def checkPrimitiveMappingHasComponent(PrimitiveToComponentMapping mapping) {
		if (mapping.target == null) {
			warning("Mapping for " + mapping.primitivecategory.name + " has no assigned component!", mapping,
				XtendUtils::primitiveToComponentMapping_Component, 0)

		}
	}

	@Check
	def checkPrimitiveConsistency(PrimitiveToComponentMapping mapping) {
		if (! (mapping.target instanceof ComponentDef))
			return;

		// 
		if (mapping.primitivecategory.primitive instanceof Component)
			return;
		var component = mapping.target as ComponentDef
		var primitive = mapping.primitivecategory.primitive
		val primitive2 = ArchitectureAbstractionUtils::resolveProxy(primitive, mapping.primitivecategory)
		var annotatedPrimitiveList = component?.annotations.filter(e|e.matches(primitive2)).filterNull
		if (annotatedPrimitiveList.size < 1) {
			warning("Component " + mapping.target.name + " is not annotated with " + mapping.primitivecategory.name,
				mapping, XtendUtils::primitiveToComponentMapping_Component, 0)
		}
	}

	@Check(CheckType::EXPENSIVE)
	def checkBounds(PatternInstance patternInstance) {
		var List<PrimitiveCategory> x = patternInstance.patternTemplate.consistsOf;
		for (PrimitiveCategory pc : x) {
			checkBounds(patternInstance, pc)
		}
	}

	def checkBounds(PatternInstance patternInstance, PrimitiveCategory pc) {
		for (c : pc.constraints) {
			checkConstraint(patternInstance, pc.primitive, c)
		}

	}

	def dispatch void checkConstraint(PatternInstance patternInstance, ComponentPrimitive primitive, Constraint c) {
	}

	//TODO check why grouping does not work when there is only one element in the group - not repeatable
	//TODO make scope context dependant -done
	//TODO check for pattern primitives that are required in a pattern but do not exist - done
	//TODO auto create pattern instances and prefill values - done
	def dispatch void checkConstraint(PatternInstance patternInstance, ComponentPrimitive primitive, Range r) {
		val isGrouped = primitive.hasGroupProperty

		var List<PrimitiveToComponentMapping> mappings = patternInstance.primitiveToComponentMappings.filter(
			e|myequals(e.primitivecategory.primitive, primitive)).toList

		// check bounds
		var targets = mappings.map(e|e.target).toList;
		var lowerBound = RangeCheckUtilities::determineLowerBound(r)
		var upperBound = RangeCheckUtilities::determineUpperBound(r)
		var numberOfFullFillingComponents = RangeCheckUtilities::calculateNumberOfFullFillingComponents(primitive.type,
			targets)
		if (!RangeCheckUtilities::checkLowerBound(primitive.type, r, targets)) {
			for (mapping : mappings) {
				var errorMessage = ""
				if (isGrouped) {
					val groupId = mapping.target.getGroupID(primitive)
					errorMessage = "Lowerbound constraint for grouped primitive " + primitive.eClass.name +
						" (GroupID: " + groupId + ") not fullfilled! #of Groups should be >= " + lowerBound + " but is " +
						numberOfFullFillingComponents
				} else {
					errorMessage = "Lowerbound constraint for primitive " + primitive.eClass.name +
						" not fullfilled! Should be >= " + lowerBound + " but is " + numberOfFullFillingComponents
				}
				error(errorMessage, patternInstance, XtendUtils::getPrimitiveToComponentMapping(),
					patternInstance.primitiveToComponentMappings.indexOf(mapping))
			}
		}
		if (!RangeCheckUtilities::checkUpperBound(primitive.type, r, targets)) {

			for (mapping : mappings) {

				var errorMessage = ""
				if (isGrouped) {
					val groupId = mapping.target.getGroupID(primitive)
					errorMessage = "Upperbound constraint for grouped primitive " + primitive.eClass.name +
						" (GroupID: " + groupId + ") not fullfilled! #of Groups should be <= " + upperBound + " but is " +
						numberOfFullFillingComponents
				} else {
					errorMessage = "Upperbound constraint for primitive " + primitive.eClass.name +
						" not fullfilled! Should be <= " + upperBound + " but is " + numberOfFullFillingComponents
				}
				error(errorMessage, patternInstance, XtendUtils::getPrimitiveToComponentMapping(),
					patternInstance.primitiveToComponentMappings.indexOf(mapping))
			}
		}

		// check group ids
		}

		/*
	 * constraint checks for layers 
	 */
		@Check(CheckType::EXPENSIVE)
		def void doCheckNoLayerIsSkipped(PatternInstance patternInstance) {
			if (!patternInstance.patternTemplate.consistsOf.exists(e|e.primitive instanceof Layering)) {
				return
			}
			val target = patternInstance.primitiveToComponentMappings.findFirst(e|e.target != null).target

			val layerAnnotations = patternInstance.primitiveToComponentMappings.filter(
				e|e.primitivecategory.primitive instanceof Layering).filterNull.toList

			// find all components per layer
			val Map<PrimitiveToComponentMapping, Set<ComponentDef>> layerComponents = new HashMap<PrimitiveToComponentMapping, Set<ComponentDef>>();
			val Map<ComponentDef, PrimitiveToComponentMapping> componentLayers = new HashMap<ComponentDef, PrimitiveToComponentMapping>();
			initializeComponentLayerMaps(layerAnnotations, target, layerComponents, componentLayers)

			/* check all layers for violations */
			for (layerAnnotation : layerAnnotations) {
				if (layerAnnotation.target != null) {
					if (layerAnnotation.target instanceof LayeringAnnotation) {
						val comps = layerComponents.get(layerAnnotation)
						val groupId = (layerAnnotation.target as LayeringAnnotation).groupId
						var boolean foundRelationToAdjacentLayer = false
						val adjacentIndex = layerAnnotations.indexOf(layerAnnotation) + 1
						
						var PrimitiveToComponentMapping targetLayer = null
						// if not at bottom layer
						if(adjacentIndex < layerAnnotations.size) {
							targetLayer = layerAnnotations.get(adjacentIndex)						
						} else {
							// do not search on the bottom layer -> fake we already found a relation
							foundRelationToAdjacentLayer = true
						}
						for (ComponentDef comp : comps) {

							/*check if a source code relation violates the relations defined in the pattern*/
							checkLayerConsistencyForComponentInSource(comp, componentLayers, layerAnnotation,
								layerAnnotations, layerComponents, groupId, patternInstance)

							/*check if the relations defined in the pattern exist in the underlying component model*/
							if (!foundRelationToAdjacentLayer) {
								
								foundRelationToAdjacentLayer = checkComponentHasRelationToAdjacentLayer(comp,
									targetLayer, componentLayers,
									layerComponents)
		
							}
						}
						
						if(!foundRelationToAdjacentLayer) {
							val String errorMessage = '''Layer «layerAnnotation.target.targetName» has no component-model relation to adjacent layer «targetLayer.target.targetName»'''
//								warning(errorMessage, layerAnnotation, InstanceDSLPackage::eINSTANCE.patternInstance_PrimitiveToComponentMappings  ,	patternInstance.primitiveToComponentMappings.indexOf(layerAnnotation))
								warning(errorMessage ,patternInstance,InstanceDSLPackage::eINSTANCE.patternInstance_PrimitiveToComponentMappings, 	patternInstance.primitiveToComponentMappings.indexOf(layerAnnotation))
						}
					}
				}
			}

			/*
		 * use stored: might be generated at the same time -- inconsistent
		 * does not "cost" 
		 * 
		 * 
		 * use evaluator to get all relations from uml class model:
		 * additional amount 
		 * has all relations - go for the uses relationships and thus might also use some relations that are not shown in the component model ( - which should actually not happen)
		 */
			}

			def private boolean checkComponentHasRelationToAdjacentLayer(ComponentDef comp,
				PrimitiveToComponentMapping adjacentLayerAnnotation,
				Map<ComponentDef, PrimitiveToComponentMapping> componentLayers,
				Map<PrimitiveToComponentMapping, Set<ComponentDef>> layerComponents) {
				val relatedComponents = comp.componentLevelDefinedDirectRelatedComponents
				val targetComps = layerComponents.get(adjacentLayerAnnotation)
				return !(Collections::disjoint(targetComps, relatedComponents))

			}

			def private checkLayerConsistencyForComponentInSource(ComponentDef comp,
				Map<ComponentDef, PrimitiveToComponentMapping> componentLayers,
				PrimitiveToComponentMapping layerAnnotation, List<PrimitiveToComponentMapping> layerAnnotations,
				Map<PrimitiveToComponentMapping, Set<ComponentDef>> layerComponents, String groupId,
				PatternInstance patternInstance) {
				val classifiers = evaluator.evaluateComponentUnwrapped(comp)

				// all layers that have a relationship and where the indizes are further apart then 1 -- all layers which have a relationship and are not adjacent
				val components = classifiers.map(c|c.relatedTargetClassifiers).flatten.components.filterNull.toSet
				val connectedLayers = components.map(f|componentLayers.get(f)).filterNull.toSet

				val inconsistentLayers = connectedLayers.filter(
					[l|(layerDistance(l, layerAnnotation, layerAnnotations) > 1)])
				for (layer : inconsistentLayers) {
					val relatedComponents = layerComponents.get(layer).filter(e|components.contains(e))
					val errorMessage = '''Layering inconsistency - Layer «groupId» (Component: «comp.name») is related to Layer «(layer.
						target as LayeringAnnotation).groupId» (Component(s): «FOR c : relatedComponents SEPARATOR ","»
				«c.name»«ENDFOR»)
			'''.toString
					warning(errorMessage, layerAnnotation, XtendUtils::primitiveToComponentMapping_Component,
						patternInstance.primitiveToComponentMappings.indexOf(layerAnnotation))
				}
			}

			def private initializeComponentLayerMaps(List<PrimitiveToComponentMapping> layerAnnotations,
				PatternInstancePrimitiveTarget target,
				Map<PrimitiveToComponentMapping, Set<ComponentDef>> layerComponents,
				Map<ComponentDef, PrimitiveToComponentMapping> componentLayers) {
				for (layerAnnotation : layerAnnotations) {
					if (layerAnnotation.target != null) {

						if (layerAnnotation.target instanceof LayeringAnnotation) {
							val layeringAnnotation = layerAnnotation.target as LayeringAnnotation
							val comps = getLayerComponents(target.eResource.contents, layeringAnnotation.groupId)
							layerComponents.put(layerAnnotation, comps)
							for (ComponentDef c : comps) {
								componentLayers.put(c, layerAnnotation)
							}
						}
					}

				}
			}

			def layerDistance(PrimitiveToComponentMapping l1, PrimitiveToComponentMapping l2,
				List<PrimitiveToComponentMapping> layerAnnotations) {

				val index1 = layerAnnotations.indexOf(l1)
				val index2 = layerAnnotations.indexOf(l2)
				return Math::abs(index1 - index2)
			}

			def components(Iterable<Classifier> classifiers) {
				val x = classifiers.map(c|c.components)
				return x;
			}

			def components(Classifier c) {
				return evaluator.getComponentForClassifier(c)
			}
		}
		